"""
dtu_conf: <DESCRIPTION>
"""
import logging

from ._version import version as __version__
from .DTU_config import DtuConfig as DtuConfig

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Fixtures for module testing."""
import io

import pytest
from appdirs import user_config_dir

from dtu_conf import DTU_config

# =============================================================================
# Example path
# =============================================================================


@pytest.fixture
def path():
    """Return the default path to use in tests."""
    # we may want the path to include authors for windows users
    return f"{user_config_dir()}/pywasp/pywasp.ini"


# =============================================================================
# Example configurations
# =============================================================================


@pytest.fixture
def initial_dtu_config(path):
    """Return an empty configuration object."""
    conf = DTU_config.DtuConfig(path)
    conf.finalise()
    return conf


@pytest.fixture
def example_dtu_config(monkeypatch, path, capsys):
    """Return a configuration with a test int and str."""
    # setup
    monkeypatch.setattr("sys.stdin", io.StringIO("\n\n"))
    conf = DTU_config.DtuConfig(path)
    conf.add_int("test", "integer", default="0")
    conf.add_str("test", "string", default="Test string")
    conf.create_config()
    conf.finalise()
    _, _ = capsys.readouterr()

    yield conf

    # teardown
    conf = DTU_config.DtuConfig(path)
    conf.finalise()


@pytest.fixture
def incomplete_config(monkeypatch, path):
    """Return a configuration with string set, required int not set."""
    # setup
    monkeypatch.setattr("sys.stdin", io.StringIO("\n" * 6))
    conf = DTU_config.DtuConfig(path)
    conf.add_int("test", "integer")
    conf.add_str("test", "string", default="Test string")
    conf.create_config()
    conf.finalise()

    yield conf

    # teardown
    conf = DTU_config.DtuConfig(path)
    conf.finalise()


# =============================================================================
# Example output
# =============================================================================


@pytest.fixture
def example_text():
    """Return example print for example_dtu_config.view()."""
    return """=== SECTIONS ===
DEFAULT <Section: DEFAULT>
test <Section: test>\n\n
=== DEFAULT ===\n
=== test ===
integer 0
string Test string"""


# =============================================================================
# List of methods
# =============================================================================


@pytest.fixture
def conf_and_methods(initial_dtu_config):
    """Return an empty configuration and its add_* method to be tested."""
    return (
        initial_dtu_config,
        [
            initial_dtu_config.add_url,
            initial_dtu_config.add_bool,
            initial_dtu_config.add_str,
            initial_dtu_config.add_int,
        ],
    )


# =============================================================================
# Example inputs
# =============================================================================


@pytest.fixture
def sections():
    """Return a list of str to use as section names."""
    return ["test"] * 6


@pytest.fixture
def keys():
    """Return a list of str to be used as key names."""
    return ["url", "bool", "string", "integer", "email", "phone"]


@pytest.fixture
def defaults():
    """Return a list of default values for the add_* methods."""
    return [
        "https://www.google.com/",
        "True",
        "Testing!",
        "0",
        "wind@domain.com",
        "+4500000000",
    ]


@pytest.fixture
def valid_indexes_and_input():
    """Return list of strings to be used as std input when creating config."""
    return (
        [1, 3, 4, 6, 8, 10],
        [
            "not an url",
            "https://www.google.com/",
            "not a bool",
            "True",
            "Every string works",
            "not an integer",
            "3",
            "not an email",
            "wind@domain.com",
            "not a phone number",
            "+4500000000",
        ],
    )



"""Tests for DTU_Config basic features."""

import io

from dtu_conf import DTU_config

# =============================================================================
# Test add methods
# =============================================================================

# class TestAddMethods:
#     """Tests related to the class method"""


def test_add_default(monkeypatch, conf_and_methods, sections, keys, defaults):
    """add_int + default create config.

    Parameters
    ----------
    methods : list of methods
    sections, keys: lists of str
        sections, keys to be used as paramters in the respective methods
    input_ : str
        Concatenated inputs to be used in respective methods. Each input should
        end with an end of line.
    defaults: list of different types(str, bool, int)
        default values to be used as paramters in the respective methods

    """
    # get configuration, ethods and std input
    conf, methods = conf_and_methods
    monkeypatch.setattr("sys.stdin", io.StringIO("\n".join(defaults)))

    # use methods
    for i, function in enumerate(methods):
        function(sections[i], keys[i], default=defaults[i])
    conf.create_config()

    # test that configuration got the required elements
    for i in range(len(methods)):
        assert conf._config.has_option(sections[i], keys[i])
        assert conf._config[sections[i]][keys[i]] == defaults[i]


def test_add(monkeypatch, conf_and_methods, sections, keys, valid_indexes_and_input):
    """add_int + create config with appropriate and inappropriate input.

    create_config should repeat prompt when an inappropriate input is used.

    Parameters
    ----------
    methods : list of methods
    sections, keys: lists of str
        sections, keys to be used as paramters in the respective methods
    input_ : str
        Concatenated inputs to be used in respective methods. Each input should
        end with an end of line.
    input_: list of different types(str, bool, int)
        default values to be used as paramters in the respective methods

    """
    # get configuration, ethods and std input
    conf, methods = conf_and_methods
    valid, input_ = valid_indexes_and_input
    monkeypatch.setattr("sys.stdin", io.StringIO("\n".join(input_)))

    # use methods
    for i, function in enumerate(methods):
        function(sections[i], keys[i])
    conf.create_config()

    # test that configuration got the required elements
    for i in range(len(methods)):
        assert conf._config.has_option(sections[i], keys[i])
        assert conf._config[sections[i]][keys[i]] == input_[valid[i]]


# =============================================================================
# Test command line tool
# =============================================================================


def test_cmd(
    example_dtu_config, sections, keys, valid_indexes_and_input, capsys, monkeypatch
):
    """Test command line for invalid and valid input"""
    # setup
    valid, input_ = valid_indexes_and_input
    i = 3
    monkeypatch.setattr("sys.stdin", io.StringIO("5"))
    _, _ = capsys.readouterr()  # ignore setup output

    # invalid input
    example_dtu_config.cmd_tool(
        args=[f"--set-{sections[i]}-{keys[i]}", input_[valid[i] - 1]]
    )
    out, err = capsys.readouterr()
    assert out.strip().split("\n")[-1] == "Please enter integer [default: 0]"

    # correction
    assert example_dtu_config._config[sections[i]][keys[i]] == "5"

    # another valid input
    example_dtu_config.cmd_tool(
        args=[f"--set-{sections[i]}-{keys[i]}", input_[valid[i]]]
    )
    assert example_dtu_config._config[sections[i]][keys[i]] == input_[valid[i]]


def test_cmd_warning(incomplete_config, recwarn):
    """Check that the command line throws an error when using non complete config"""
    incomplete_config.cmd_tool(args=["--set-test-string", "wind"])
    assert len(recwarn) == 1
    w = recwarn.pop()
    assert str(w.message) == "Option integer of section test is required and not set."


# =============================================================================
# Other tests
# =============================================================================


def test_str(example_dtu_config, example_text, capsys):
    """view method should print the sections, options, values."""
    # ignore setup output
    _, _ = capsys.readouterr()
    # print configuration and check std out
    print(example_dtu_config)
    out, err = capsys.readouterr()
    assert out.strip() == example_text


def test_get(incomplete_config, example_dtu_config, capsys, recwarn):
    incomplete_config.get_option("test", "integer")
    out, err = capsys.readouterr()
    assert out == ""

    example_dtu_config.get_option("test", "option_that_does_not_exists")
    assert len(recwarn) == 1

    example_dtu_config.get_option("test", "integer") == "0"


def test_file_integrity(example_dtu_config, path):
    """
    Check that creating an object does not corrupt the existing configuration file

    We might want to check the same for other actions.
    """
    # create file
    example_dtu_config.save()
    text1 = example_dtu_config._read_config()

    # create new config object at the same path
    config2 = DTU_config.DtuConfig(path)
    # we haven't finalised the config yet

    # check that the new object read the file rather than erasing it.
    assert text1 == config2._read_config()



"""Configuration class."""

import argparse
import os
import warnings
from configparser import ConfigParser
from pathlib import Path
from urllib.error import URLError
from urllib.request import urlopen

import phonenumbers
from appdirs import user_config_dir
from pydantic import AnyHttpUrl, EmailStr, validate_arguments

# =============================================================================
# Help functions to check the input types and validity
# =============================================================================


@validate_arguments
def _is_pydantic_url(value: AnyHttpUrl):
    """Use function.validate(arg) to check user input."""
    pass


@validate_arguments
def _is_pydantic_bool(value: bool):
    """Use function.validate(arg) to check user input."""
    pass


@validate_arguments
def _is_pydantic_str(value: str):
    """Use function.validate(arg) to check user input."""
    pass


@validate_arguments
def _is_pydantic_int(value: int):
    """Use function.validate(arg) to check user input."""
    pass


@validate_arguments()
def _is_pydantic_email(value: EmailStr):
    """Use function.validate(arg) to check user input."""
    pass


def _get_valid_input(
    validator, prompt_text, value=None, default=None, limit=5, required=True
):
    """
    Ask for input until the value pass the validator or limit of tries exeeded.

    Parameters
    ----------
    validator : function
        Returns True is the input is adapted.
    prompt_text : str
        Text to prompt the user for another input.
    value : str, int, or bool, optional.
        Initial proposition for the value. The default is None.
    default: str, int or bool
        default proposition for the value, used if user enters nothing.
    limit : int, optional
        Number of times to ask for input before throwing an error. The default is 5.
    required : bool, optional
        Whether the value is required from the user.

    Returns
    -------
    value : str, int, or bool
        value that pass the validator.

    """

    i = 0
    while i < limit:
        try:
            # start by asking value unless we get user input the first time
            if (i > 0) or value is None:
                value = input(prompt_text).strip() or default
            validator(value)
            return value
        except Exception as e:
            i += 1
            # if value is not required, accept None as an answer after first input
            if (value is None) and (not required):
                return None
            print(e)
            if not required:
                print("Press enter to skip.")


def _check_url_working(url):
    """Validate that the requested url is responding.

    Parameters
    ----------
    url : string
        Path to the license server, including port
    """
    check_text = """Please check that you have entered the url correctly. If this issue persists contact your local license server administrator."""
    try:
        res = urlopen(url)
    except URLError:
        raise ConnectionError(f"Could not connect to {url}. {check_text}")

    if res.status != 200:
        raise ConnectionError(f"Could not connect to {url}. {check_text}")


# =============================================================================
# Main class
# =============================================================================


class DtuConfig:
    """
    Class for creation and manipulation of the application configuration file.

    Parameters
    ----------
    path : str or Path, optional
        Path to the application configuration file.
        Using app name and author is recommended for portability but path will
        have precedence if given.
    app_name: str, optional
        Name of the application
    app_author: str, optional
        Author of the application. Used for Windows path.
    """

    def __init__(self, path=None, app_name=None, app_author=None):
        """See class documentation."""
        # check that we have enough information for the path
        if not (path or app_name and app_author):
            raise Exception(
                "Cannot create configuration: parameters should include the path of the configuration or the application name and author."
            )
        # name/author will be used only if path is not given.
        if not path:
            path = f"{user_config_dir(app_name, app_author)}/{app_name}.ini"
        self.path = Path(path)
        self._options = {}
        self._config = ConfigParser()
        if self.path.exists():
            self._config.read(self.path)

    # =========================================================================
    # add options to the object
    # =========================================================================

    def _add_option(self, section, option, **kwargs):

        kwargs["is_set"] = self._config.has_option(section, option)

        if section not in self._options.keys():
            self._options[section] = {option: kwargs}
        else:
            self._options[section][option] = kwargs

    def _add_general(
        self,
        section,
        option,
        pydantic_type,
        input_text,
        default,
        required,
        validator,
    ):
        if input_text is None:
            input_text = f"Please enter {option}"
            if not required:
                input_text = f"{input_text} (optional)"
        if default is not None:
            input_text = f"{input_text} [default: {default}]"
        input_text = f"{input_text} \n"

        self._add_option(
            section,
            option,
            pydantic_type=pydantic_type,
            input_text=input_text,
            default=default,
            required=required,
            validator=validator,
        )

    def add_url(
        self,
        section,
        option,
        pydantic_type=AnyHttpUrl,
        input_text="Please enter a valid URL.",
        default=None,
        required=True,
    ):
        """
        Add a url to the configuration.

        By default the method checks that the url is valid. This will fail if
        the user is offline.

        Parameters
        ----------
        section: str
            String representing the configuration file section
        option : str
            Option of the configuration section that is being updated
        value : int, float or str
            Parameter value
        input_text : str
            prompt
        """

        def validator(value):
            _is_pydantic_url.validate(value)
            _check_url_working(value)

        self._add_general(
            section, option, pydantic_type, input_text, default, required, validator
        )

    def add_bool(
        self,
        section,
        option,
        pydantic_type=bool,
        input_text=None,
        default=None,
        required=True,
    ):
        """
        Add a boolean parameter to the configuration.

        Parameters
        ----------
        section: str
            String representing the configuration file section
        option : str
            Option of the configuration section that is being updated
        value : int, float or str
            Parameter value
        input_text : str
            prompt. If None, a default value will be created: "Please enter
            {option}"
        """

        def validator(value):
            _is_pydantic_bool.validate(value)

        self._add_general(
            section, option, pydantic_type, input_text, default, required, validator
        )

    def add_str(
        self,
        section,
        option,
        pydantic_type=str,
        input_text=None,
        default=None,
        required=True,
    ):
        """
        Add a string parameter to the configuration.

        Parameters
        ----------
        section: str
            String representing the configuration file section
        option : str
            Option of the configuration section that is being updated
        value : int, float or str
            Parameter value
        input_text : str, optional
            prompt. If None, a default value will be created: "Please enter
            your {option}"
        """
        if input_text is None:
            input_text = f"Please enter your {option}"
            if not required:
                input_text = f"{input_text} (optional)"

        def validator(value):
            _is_pydantic_str.validate(value)

        self._add_general(
            section, option, pydantic_type, input_text, default, required, validator
        )

    def add_int(
        self,
        section,
        option,
        pydantic_type=int,
        input_text=None,
        default=None,
        required=True,
    ):
        """
        Add an integer parameter to the configuration.

        To add constraints on the value please define a custom validator.

        Parameters
        ----------
        section: str
            String representing the configuration file section
        option : str
            Option of the configuration section that is being updated
        value : int, float or str
            Parameter value
        input_text : str
            prompt. If None, a default value will be created: "Please enter
            {option}"
        """

        def validator(value):
            _is_pydantic_int.validate(value)

        self._add_general(
            section, option, pydantic_type, input_text, default, required, validator
        )

    def add_email(
        self,
        section,
        option,
        pydantic_type=EmailStr,
        input_text="Please enter your email adress.",
        default=None,
        required=True,
    ):
        """
        Add an email to the configuration.

        The method checks that the email has a valid format but not if it
        exists. To do so, send a confirmation email to the user.

        Parameters
        ----------
        section: str
            String representing the configuration file section
        option : str
            Option of the configuration section that is being updated
        value : int, float or str
            Parameter value
        input_text : str
            prompt
        """

        def validator(value):
            _is_pydantic_email.validate(value)

        self._add_general(
            section, option, pydantic_type, input_text, default, required, validator
        )

    def add_phone_number(
        self,
        section,
        option,
        pydantic_type="phone_number",
        input_text="Please enter your phone number.",
        default=None,
        required=True,
    ):
        """
        Add a phone number to the configuration.

        By default the method checks if the number is valid. country code is
        required. To change requirements please define a default validator.

        Parameters
        ----------
        section: str
            String representing the configuration file section
        option : str
            Option of the configuration section that is being updated
        value : int, float or str
            Parameter value
        input_text : str
            prompt
        """

        def validator(value):
            number = phonenumbers.parse(value)

        self._add_general(
            section, option, pydantic_type, input_text, default, required, validator
        )

    # =========================================================================
    # create and edit the configuration
    # =========================================================================

    def create_config(self):
        """
        Create configuration.

        The user will be asked the values od the parameters that have not been
        saved.
        """
        conf = self._config

        # Loop over all options in the dataset
        for sec, opt_dict in self._options.items():
            if not conf.has_section(sec):
                conf.add_section(sec)
            for opt, meta in opt_dict.items():
                if not meta["is_set"]:
                    # Get user input and run checks:
                    # Pydantic & additional checks like URL is reachable
                    try:
                        val = _get_valid_input(
                            meta["validator"],
                            meta["input_text"],
                            default=meta["default"],
                            required=meta["required"],
                        )
                        if val:
                            conf[sec][opt] = str(val)
                            meta["is_set"] = True
                    except Exception as e:
                        print(e)
        self.save()
        return conf

    def save(self):
        """Save the configuration file in the application directory."""
        if not os.path.isdir(os.path.dirname(self.path)):
            os.makedirs(os.path.dirname(self.path))
        with open(self.path, "w") as config_file:
            self._config.write(config_file)

    def _read_config(self):
        """Load configuration file to DtuConfig class."""
        self._config.read(self.path)

    def finalise(self):
        """
        Remove _config sections/options that are not defined in _options.

        The configuration file is updated with the current options and saved.
        """
        for section in self._config.sections():
            # delete unused sections
            if section not in self._options:
                self._config.remove_section(section)
            # delete unused options
            else:
                for option in self._config[section]:
                    if option not in self._options[section]:
                        self._config.remove_option(section, option)
        self.save()

    # =========================================================================
    # parser and command line tool to edit the configuration
    # =========================================================================

    def _check_required_options(self):
        """Throw warnings if required options are not set."""
        for sec, opt_dict in self._options.items():
            for key in self._options[sec].keys():
                meta = opt_dict[key]
                if meta["required"] and not meta["is_set"]:
                    warnings.warn(
                        f"Option {key} of section {sec} is required and not set."
                    )

    def _create_parser(self):
        """Build ArgumentParser using settings of config object."""
        parser = argparse.ArgumentParser(
            description="Update configuration file", conflict_handler="resolve"
        )

        for section in self._options.keys():
            for key in self._options[section].keys():
                parser.add_argument(
                    f"--set-{section}-{key}",
                    help=f"Pass a value to section {section}, key {key}",
                )

        return parser.parse_args

    def cmd_tool(self, args=None):
        """Use argparse to values from the user."""
        pars = self._create_parser()
        if args:
            args = pars(args)
        else:
            args = pars()

        # process arguments from the user
        conf = self._config

        for sec, opt_dict in self._options.items():
            for key in self._options[sec].keys():
                if args.__dict__[f"set_{sec}_{key}"] is not None:
                    if not conf.has_section(sec):
                        conf.add_section(sec)
                    meta = opt_dict[key]
                    # Pydantic & additional checks like URL is reachable
                    try:
                        val = _get_valid_input(
                            meta["validator"],
                            meta["input_text"],
                            value=args.__dict__[f"set_{sec}_{key}"],
                            required=meta["required"],
                            default=meta["default"],
                        )
                        if val:
                            conf[sec][key] = str(val)
                            meta["is_set"] = True
                    except Exception as e:
                        print(e)

        # Throw warning if required options are not set
        self._check_required_options()

        # Save config to disk
        self.save()

    # =========================================================================
    # get options from the object, view the content of the configuration
    # =========================================================================

    def get_option(self, section, option):
        """Return the value of the option if it is set or None."""
        if section not in self._options:
            warnings.warn(f"The section {section} does not exists.")
        if option not in self._options[section]:
            warnings.warn(f"The option {option} does not exists in section {section}")

        try:
            return self._config[section][option]
        except KeyError:
            return None

    def __str__(self):
        """Return configuration contents."""
        separator1 = "\n==="
        separator2 = "===\n"
        out = ""
        out += f"{separator1} SECTIONS {separator2}"
        for k, v in self._config.items():
            out += f"{k} {v}\n"
        out += "\n"
        for section, _ in self._config.items():
            out += f"{separator1} {section} {separator2}"
            # print keys, values
            for k, v in self._config.items(section):
                out += f"{k} {v}\n"
        return out

    def __repr__(self):
        """Return configuration contents."""
        return self.__str__()

    def show_options(self):
        """Print options dictionary."""
        separator1 = "\n==="
        separator2 = "===\n"
        out = ""
        out += f"{separator1} SECTIONS {separator2}"
        for sect in self._options:
            out += f"{sect}\n"
        for section in self._options:
            out += f"{separator1} {section} {separator2}"
            # print keys, values
            for opt in self._options[section]:
                out += f"{opt} {self._options[sect][opt]['pydantic_type']}\n"
        print(out)


if __name__ == "__main__":
    app_name = "pywasp"  # to modify
    app_author = "author"  # to modify

    # creation of object. If a file exists, it's loaded
    conf = DtuConfig(app_name=app_name, app_author=app_author)

    # add options. If  they have been set previously the user won't be asked
    # the values again.
    conf.add_int("test", "integer", default="0")
    conf.add_str("test", "string", default="Test string")
    conf.create_config()
    conf.finalise()

    # command interface
    conf.cmd_tool(args=["--set-test-string", "hi"])

    # print(conf)
    conf.show_options()



"""Configuration of additional command line arguments for pytest."""
# import importlib


# imports fixtures
from fixtures.fixture_basic import (
    conf_and_methods,
    defaults,
    example_dtu_config,
    example_text,
    incomplete_config,
    initial_dtu_config,
    keys,
    path,
    sections,
    valid_indexes_and_input,
)
